#ifndef MLN_KERNEL_EXECUTE_HPP
# define MLN_KERNEL_EXECUTE_HPP

# include <mln/core/neighborhood/neighborhood.hpp>
# include <mln/core/image/image.hpp>
# include <mln/kernel/context.hpp>

namespace mln
{

  namespace kernel
  {

    ///
    /// \brief Execute a kernel expression
    ///
    ///
    template <class Expr, class N, class... I>
    void
    execute(const Expr& expr, const Neighborhood<N>& nbh, I&&... images);



    /******************************************/
    /****          Implementation          ****/
    /******************************************/

    namespace impl
    {

      template <class Expr, class Nbh, class... I>
      void
      execute(Expr& expr, const Nbh& nbh, I&&... images)
      {
        kernel::kernel_context<Expr, Nbh, I...> ctx(nbh, std::forward<I>(images)...);

        kernel::details::transform_aggregate t1;
        kernel::details::retrieve_aggregate  t2;

        auto newexpr = t1(expr);
        auto subexprs = t2(expr);

        for (ctx.p_init(); !ctx.p_finished(); ctx.p_next())
	  {
	    // Pre-computation of aggregates
	    for (ctx.n_init(); !ctx.n_finished(); ctx.n_next())
	      {
		
	      }

          proto::eval(expr, ctx);
	  }
      }

    }


    template <class Expr, class N, class... I>
    void
    execute(const Expr& expr, const Neighborhood<N>& nbh, I&&... images)
    {
      impl::execute(expr, exact(nbh), std::forward<I>(images)...);
    }




  }

} // end of namespace mln::kernel

#endif //!MLN_KERNEL_EXECUTE_HPP
